import ut
from ctypes import POINTER
from comtypes.client import CoGetObject

# WMI has dual interfaces.
# Some methods/properties have "[out] POINTER(VARIANT)" parameters.
# This test checks that these parameters are returned as strings:
# that's what VARIANT.__ctypes_from_outparam__ does.
class Test(ut.TestCase):
    def test_wmi(self):
        wmi = CoGetObject("winmgmts:")
        disks = wmi.InstancesOf("Win32_LogicalDisk")

        # There are different WMI typelibs installed on win2k and
        # winXP.  This has the unfortunate effect that we cannot use a
        # named wrapper.  So, for py2exe (if we want to use it) we
        # have to import the wrappers with their fully spelled out name:
##        try:
##            import comtypes.gen._565783C6_CB41_11D1_8B02_00600806D9B6_0_1_1 as mod
##        except ImportError:
##            import comtypes.gen._565783C6_CB41_11D1_8B02_00600806D9B6_0_1_2 as mod

        # If we are not using py2exe and don't want to spell out the
        # typelib wrappers name as above, how would we access the module?
        #
        # Very obscure.  Should we have a utiliy functions 'GetTypelibWrapper(obj)' ?
        mod = __import__(wmi.__com_interface__.__module__, globals(), locals(), ['dummy'])
        mod.wbemPrivilegeCreateToken

        for item in disks:
            # obj[index] is forwarded to obj.Item(index)
            # .Value is a property with "[out] POINTER(VARIANT)" parameter.
            a = item.Properties_["Caption"].Value
            b = item.Properties_.Item("Caption").Value
            self.failUnlessEqual(a, b)
            self.failUnless(isinstance(a, basestring))
            self.failUnless(isinstance(b, basestring))
##            for prop in item.Properties_:
##                    print "\t", (prop.Name, prop.Value)
            if __debug__:
                print a, item.Properties_["Description"].Value
        # len(obj) is forwared to obj.Count
        c1 = len(disks)
        c2 = disks.Count
        self.failUnlessEqual(c1, c2)

if __name__ == "__main__":
    ut.main()
