import unittest, os
from ctypes import *
from ctypes.com import IUnknown, GUID
from ctypes.com.automation import VARIANT, LoadTypeLibEx, \
     VT_ARRAY, VT_VARIANT, VT_I4, VT_INT, oleaut32, IDispatch

# test VT_ARRAY in VARIANTS
class SAFEARRAYTest(unittest.TestCase):

    def test_VT_VARIANT(self):
        v = VARIANT()
        # lists...
        v.value = [1, "spam", [None, "foo"]]
        self.failUnlessEqual(v.value, [1, "spam", [None, "foo"]])
        self.failUnlessEqual(v.vt, VT_VARIANT | VT_ARRAY)
        dim = oleaut32.SafeArrayGetDim(v._.voidp)
        self.failUnlessEqual(dim, 1)
        # and tuples.
        v.value = (1, "spam", (None, "foo"))
        self.failUnlessEqual(v.value, [1, "spam", [None, "foo"]])

    def test_VT_INT(self):
        import array
        data = array.array("i", (1, 2, 3, 4, 5))
        v = VARIANT()
        v.value = data
        self.failUnlessEqual(v.vt, VT_INT | VT_ARRAY)
        dim = oleaut32.SafeArrayGetDim(v._.voidp)
        self.failUnlessEqual(dim, 1)

    def _getrefcnt(self, comobj):
        comobj.AddRef()
        return comobj.Release()

    def test_VT_UNKNOWN(self):
        tlb = LoadTypeLibEx(r"c:\windows\system32\shdocvw.dll")
        self.failUnlessEqual(self._getrefcnt(tlb), 1)
        v = VARIANT()
        v.value = tlb
        self.failUnlessEqual(self._getrefcnt(tlb), 2)
        v.value = None
        self.failUnlessEqual(self._getrefcnt(tlb), 1)

        v.value = [tlb]
        self.failUnlessEqual(self._getrefcnt(tlb), 2)
        pcom = v.value[0]
        self.failUnlessEqual(self._getrefcnt(tlb), 3)
        self.failUnlessEqual(self._getrefcnt(pcom), 3)

        # ITypeLib derives from IUnknown, and VARIANT can only store IUnknown or IDispatch
        # pointers, so:
        self.failUnlessEqual(type(pcom), POINTER(IUnknown))

        del pcom

        self.failUnlessEqual(self._getrefcnt(tlb), 2)
        v.value = None
        self.failUnlessEqual(self._getrefcnt(tlb), 1)

    def test_VT_DISPATCH(self):
        from ctypes.com.client import Dispatch
        p = Dispatch("MSScriptControl.ScriptControl") # any inproc IDispatch will do
        pdisp = p._comobj

        self.failUnlessEqual(self._getrefcnt(pdisp), 1)
        v = VARIANT()
        v.value = pdisp
        self.failUnlessEqual(self._getrefcnt(pdisp), 2)
        v.value = None
        self.failUnlessEqual(self._getrefcnt(pdisp), 1)

        v.value = [pdisp]
        self.failUnlessEqual(self._getrefcnt(pdisp), 2)
        pcom = v.value[0]
        self.failUnlessEqual(self._getrefcnt(pdisp), 3)
        self.failUnlessEqual(self._getrefcnt(pcom), 3)

        # ITypeLib derives from IUnknown, and VARIANT can only store IUnknown or IDispatch
        # pointers, so:
        self.failUnlessEqual(type(pcom), POINTER(IDispatch))

        del pcom

        self.failUnlessEqual(self._getrefcnt(pdisp), 2)
        v.value = None
        self.failUnlessEqual(self._getrefcnt(pdisp), 1)

##    def test_multidim()
        
if __name__ == "__main__":
    unittest.main()
